import { append, printable, spliterate, stringAndSymbolicEntriesOf, throwParseError } from "@arktype/util";
import { Scanner } from "./string/shift/scanner.js";
export const parseObjectLiteral = (def, ctx) => {
    let spread;
    const structure = {};
    // We only allow a spread operator to be used as the first key in an object
    // because to match JS behavior any keys before the spread are overwritten
    // by the values in the target object, so there'd be no useful purpose in having it
    // anywhere except for the beginning.
    const parsedEntries = stringAndSymbolicEntriesOf(def).map(parseEntry);
    if (parsedEntries[0]?.kind === "spread") {
        // remove the spread entry so we can iterate over the remaining entries
        // expecting non-spread entries
        const spreadEntry = parsedEntries.shift();
        const spreadNode = ctx.$.parse(spreadEntry.value, ctx);
        if (!spreadNode.hasKind("intersection") || !spreadNode.structure) {
            return throwParseError(writeInvalidSpreadTypeMessage(typeof spreadEntry.value === "string" ?
                spreadEntry.value
                : printable(spreadEntry.value)));
        }
        spread = spreadNode.structure;
    }
    for (const entry of parsedEntries) {
        if (entry.kind === "spread")
            return throwParseError(nonLeadingSpreadError);
        if (entry.kind === "index") {
            // handle key parsing first to match type behavior
            const key = ctx.$.parse(entry.inner, ctx);
            const value = ctx.$.parse(entry.value, ctx);
            // extract enumerable named props from the index signature
            const [enumerable, nonEnumerable] = spliterate(key.branches, (k) => k.hasKind("unit"));
            if (enumerable.length) {
                structure.required = append(structure.required, enumerable.map(k => ctx.$.node("required", { key: k.unit, value })));
                if (nonEnumerable.length) {
                    structure.index = append(structure.index, ctx.$.node("index", { index: nonEnumerable, value }));
                }
            }
            else {
                structure.index = append(structure.index, ctx.$.node("index", { index: key, value }));
            }
        }
        else {
            const value = ctx.$.parse(entry.value, ctx);
            structure[entry.kind] = append(structure[entry.kind], {
                key: entry.inner,
                value
            });
        }
    }
    const structureNode = ctx.$.node("structure", structure);
    return ctx.$.schema({
        domain: "object",
        structure: spread?.merge(structureNode) ?? structureNode
    });
};
export const nonLeadingSpreadError = "Spread operator may only be used as the first key in an object";
export const parseEntry = (entry) => Object.assign(parseKey(entry[0]), { value: entry[1] });
const parseKey = (key) => typeof key === "symbol" ? { inner: key, kind: "required" }
    : key.at(-1) === "?" ?
        key.at(-2) === Scanner.escapeToken ?
            { inner: `${key.slice(0, -2)}?`, kind: "required" }
            : {
                inner: key.slice(0, -1),
                kind: "optional"
            }
        : key[0] === "[" && key.at(-1) === "]" ?
            { inner: key.slice(1, -1), kind: "index" }
            : key[0] === Scanner.escapeToken && key[1] === "[" && key.at(-1) === "]" ?
                { inner: key.slice(1), kind: "required" }
                : key === "..." ? { inner: "...", kind: "spread" }
                    : { inner: key === "\\..." ? "..." : key, kind: "required" };
export const writeInvalidSpreadTypeMessage = (def) => `Spread operand must resolve to an object literal type (was ${def})`;
